function [xBest, fBest,x]=RCO(func,D0,FEmax,xInit) % Ruler and Compass Optimisation
%{
    Maurice.Clerc@WriteMe.com
   
A simple deterministic optimisation method
(but optionnaly pseudo-stochastic).


2024-06-10
Main modifications:
- A DEFAULT set of options.
- You can specify xInit, a set of 2^D positions of D components.
This set defines a subspace of the definition space, but even if it doesn't
contain the solution RCO may escape and find the global minimum.
Try 
RCO(4,1,15,[0;4];
xInit=[-2,-1.5;-0.5,-2;-0.5,-0.5;-1.5,-0.5];RCO(10,2,50,xInit);

- A "cheap" option.

Main conclusions
----------------
Pro:
- On some functions supposed to be difficult may find a good solution 
  after just a few evaluations. Interesting if each evaluation
  is very costly.

Neutral:
- Sometimes (rarely) even  "wrong" lowerbound may be better than a correct one.
  Try for example Six Hump Camel Back. In get_func.m set the lowerbound
  to 0 instead of -1.1.

Con:
- Only for low dimension (typically <10), at least for this version.

- Cant't find a solution on the boundary. But you can "artificially"
  extend the definition space thanks to fMax (see function Eggholder +)

- Doesn't work well on (partly) discrete problems

- May stagnate in a local minimum if its value is very close to
  the one of the global minimum.

 - The "cheap" method is not very good and even sometimes very bad.

Examples
RCO(4,1,15) % Bi-parabola
RCO(10,2,50) % Six Hump Camel Back
RCO(9,3,2000) %  Rosenbrock
RCO(1,4,50000) % Rastrigin
RCO(49,10,50000) % Lévy

%}

warning('off') % Don't display warnings (in particular for singular matrix)
rng(123456789) % For reproducibility of test problems that use rand
global noWayCount % At time t no intersection can be computed
global outsideCount %  At time t intersection is outside the search space
global FE
 % Number of steps
FE=0;
noWayCount=zeros(1,FEmax); 
outsideCount=zeros(1,FEmax); 
global exploit
exploit=zeros(1,FEmax);
global fMax
global savePos fid

DEFAULT=true; 

if DEFAULT
    adaptOption= 0;
    stagDetect=false;
    out2inOption=0;
    cheap=0;
    savePos=false;  % Save positions for further analysis
else % You may play with different options, to experiment
adaptLBcoeff=0.5; % 0.5; % In ]0,1[
adaptOption=3; % 0, 1, 2,  3 
                %{ 
                    Adaptation method. 
                     If >0 RCO is parameter free. No need to define
                     a lower bound.
                    It may be useful to try all.
                    Note: with adaptLBcoeff=0.5 
                     adaptOption 1 and 3 are the same 
                    If adaptLBcoeff>0.5 then 3 seems better
                %}
stagDetect=false; % If true: stagnation detection is used and
                    %  and a new point is defined. Experimental.
out2inOption=0; % <0 => use pseudo randomness 
                % Might be interesting for (partly) discrete problems
                % as Pressure Vessel. Experimental.
                
 cheap=0;  % >0 => lower complexity and faster methods
                  % Usually worse results, but sometimes better.
                  % Experimental. You shouldn't use it for D=1
 savePos=true;
end
global x0 % Set of corners
global toCorner % Number of times a corner is used
                % to construct a new point (except at the very beginning)
toCorner=0;

% Possible plots for D=1 or 2
plotLandscape=true;
plotLine=true;
plotPoints=false;

% Possible plots for any problem
plotNowayOutside=false;
plotBalance=false;

%-------------------------------
if savePos
    fid = fopen('positions', 'w');
end

fprintf('\n=============================================================')
% Define the problem
[xMin, xMax, lowerBound,quantis,D]=get_func(func,D0);

fprintf('\n function %i, dimension %i, budget %i',func,D,FEmax);

tStart=cputime; % Start for computing time
% -------------------------------- Initialisation
% Define all "corners" (2^D)
if exist('xInit')
  x=xInit;
else
    x=corners(xMin,xMax);
end
x0=x; % May be useful for out2in


% Evaluate them
[nCorners,~]=size(x);
for n=1:nCorners
  f(n)=fit(x(n,:),func,0);
end

fMax=max(f); % Useful to "extend" a definition space

% Save the best
[fBest,Ind]=min(f);
fprintf('\n Best after init %e',fBest)
xBest=x(Ind,:);
fBestQ=fBest; % For (partly) discrete problems

plotRun=(plotPoints || plotLine) && D<=2;
if plotRun
  for n=1:nCorners
    line(n,1:D)=x(n,:); line(n,D+1)=f(n);
  end
end

% --------------------------------
if cheap
  %pairs=nchoosek(1:nCorners,2); % All possible pairs
  for n=1:nCorners-1 pairs(n,1)=n; pairs(n,2)=n+1; end % Sequential pairs
 %k=1; for n=1:2^(D-1) pairs(n,1)=k; pairs(n,2)=k+1; k=k+2; end

else 
  pairs=[];
end

if out2inOption<0  || stagDetect % Pseudo randomness
  magicPrime(FEmax*D);
end


noImprov=0; % Number of improvements
iter=0;
while FE<FEmax
  %fprintf('\n t %i, fBest %e',t,fBest)
    if ~isempty(quantis)
      fB=fBestQ;
    else
      fB=fBest;
    end
    
    switch adaptOption
      case 1 % Method 1
        if fB>=0
          lb=adaptLBcoeff*fB;
        else
          lb=(1+adaptLBcoeff)*fB;
        end
        
      case 2 % Method 2
        lb=adaptLBcoeff*lowerBound +(1-adaptLBcoeff)*fB;
        
      case 3 % Method 3
        if fB>0
          lb=fB*adaptLBcoeff;
        else
            lb=fB*(2-adaptLBcoeff);
        end
  
    %end
      otherwise % User defined lower bound
    lb=lowerBound;
    end
  
 
 if stagDetect
  if noImprov>nCorners  % If no improvement during a cycle
    % try another method in case of intersection
    out2inOption=-2; % try 1;
    noImprov=0;
  end
 end
 %fprintf('\n %i',out2inOption)
  % Define a new position
  [xn,fn]=new(x,f,xMin,xMax,func,lb,out2inOption,cheap,pairs);
  iter=iter+1;
  LB(iter)=lb; % Useful for plot
  %fprintf('\n t %i, lb %e',t,lb)
  
  if plotRun % Save positions for future plots
    line(iter+nCorners,1:D)=xn;
    line(iter+nCorners,D+1)=fn;
  end
  
  % For (partly) discrete problems
  if ~isempty(quantis)
    if fn<fBestQ % Keep the best solution after discretisation
      xQ=quantify(xn,quantis);
      fQ=fit(xQ,func);
      if fQ<fBestQ
        fBestQ=fQ;
        xBest=xQ; % LB=zeros(iter+nCorners,D+1); % Just to save the lower bounds.

      else
        noImprov=noImprov+1;
      end
    end
  else
    if fn<fBest
      fBest=fn;
      xBest=xn;
    else
      noImprov=noImprov+1;
    end
  end
  
  % Define another line (D=1) or plane (D=2) or hyperplane (D>2)
  % (treadmill method)
  for p=1:nCorners-1
    x(p,:)=x(p+1,:);
    f(p)=f(p+1);
  end
  x(nCorners,:)=xn;
  f(nCorners)=fn;
end % end of the run

tEnd=cputime; % Stop computing time
% -------------------------------- Results
if ~isempty(quantis)  % Some problems may be (partly) discrete
    fBest=fBestQ;
end
fprintf('\n Function %i',func);
fprintf('\n out2inOption %i',out2inOption);
if cheap fprintf('\n cheap option'); end
if stagDetect
  fprintf('\n Stagnation detection activated');
end
  switch adaptOption
    case 1
      fprintf('\n Adaptation 1')
      fprintf('\n Totally adaptive lower bound, coefficient %f',adaptLBcoeff);
    case 2
      fprintf('\n Adaptation 2');
      fprintf('\n Mix user defined lower bound %e + adaptive,  coefficient %f',...
        lowerBound,adaptLBcoeff);
      fprintf('\n (no matter the sign of fBest)')
    case 3
      fprintf('\n Adaptation 3');
       fprintf('\n Totally adaptive lower bound, coefficient %f',adaptLBcoeff);

    case 4
      fprintf('\n Adaptation 4, Lowerbound=fBest'); % BAD
  otherwise
  fprintf('\n lowerBound %f',lowerBound);
  end

fprintf('\n Evaluations %i', FE);
fprintf('\n noWayCount %i',sum(noWayCount));
fprintf('\n outsideCount %i',-sum(outsideCount));
fprintf('\n toCorner %i',toCorner);
fprintf('\n final fMax %e',fMax)

fprintf('\n fBest %e',fBest)
fprintf('\n at \n'); disp(xBest)

fprintf('\n CPU time %e \n',tEnd-tStart)
% ================================== Plots
savePos=false;
if plotRun 
 % LB=zeros(iter+nCorners,D+1); % Just to save the lower bounds.
  figure; 
end

if plotLandscape && D<=2
  fPlot(func,D,xMin,xMax,lowerBound);
  hold on
end

if plotLine
  switch D
    case 1
      plot(line(:,1), line(:,2),'r-','LineWidth',2);
      hold on
    case 2
      plot3(line(:,1), line(:,2),line(:,3),'r-','LineWidth',2);
      hold on
  end
end

if plotPoints
  switch D
    case 1
      plot(line(:,1), line(:,2),'r.','MarkerSize',12)
      %axis([xMin,xMax,lowerBound, max(line(:,2))]);
      hold on
    case 2
      plot3(line(:,1), line(:,2),line(:,3),'r.','MarkerSize',12)
      hold on
  end
end


if plotNowayOutside
  figure
  bar(noWayCount,0.1, 'k');
  hold on
  bar(outsideCount,0.1,'k');
  xlabel("Evaluations");
  if sum(noWayCount)>0
   label_h = ylabel('outside | noway');
  else
    label_h = ylabel('outside');
  end
   % label_h.Position(2) = 0; % change vertical position of ylabel
    set(gca, 'YTick', []);  
end

if plotBalance % Normally associated to plotNowayOutside
   balance(1)=0; % Exploration/exploitation
   for fe=2:FEmax
     exploit(fe)=exploit(fe-1)+ exploit(fe);
     balance(fe)=(fe-exploit(fe))/exploit(fe);
   end
   if ~plotNowayOutside
     figure
   else
     yyaxis right
   end
   plot(balance(2^D:FEmax),"Color","r", 'LineWidth',2)
   xlabel("Evaluations");
   %yyaxis right
   ylabel("Exploration/exploitation", "Color", "r")
end
end
 



